/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.remote;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.Key;
import dk.vajhoej.record.RecordException;

/**
 * Class ByteClient implement IsamProtocolClient using byte arrays
 * over plain Stream's.
 */
public class ByteClient implements IsamProtocolClient {
    private static final boolean USEBUFFER = true;
    private static Logger log = Logger.getLogger(ByteClient.class.getName());
    private InputStream is;
    private OutputStream os;
    /**
     * Create instance of ObjectClient.
     * @param is input stream
     * @param os output stream
     */
    public ByteClient(InputStream is, OutputStream os) {
        if(USEBUFFER) {
            this.is = new BufferedInputStream(is);
            this.os = new BufferedOutputStream(os);
        } else {
            this.is = is;
            this.os = os;
        }
        log.finest("Streams open");
    }
    public void handshake() throws IsamException {
        try {
            os.write(IsamProtocol.B_V1);
            os.flush();
            log.finer("Handshake complete");
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    @SuppressWarnings("unchecked")
    public <T, TK> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Read class " + t.getName() + " with key " + k);
        try {
            os.write(IsamProtocol.READ);
            ByteHelper.writeClass(os, t);
            ByteHelper.writeKey(os, k);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
            return (T)ByteHelper.readObject(is, log);
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T> void create(T o) throws IsamException, RecordException {
        log.fine("Create class " + o.getClass().getName() + " with value " + o);
        try {
            os.write(IsamProtocol.CREATE);
            ByteHelper.writeObject(os, log, o);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T> void update(T o) throws IsamException, RecordException {
        log.fine("Update class " + o.getClass().getName() + " with value " + o);
        try {
            os.write(IsamProtocol.UPDATE);
            ByteHelper.writeObject(os, log, o);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T, TK> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Delete class " + t.getName() + " with key " + k);
        try {
            os.write(IsamProtocol.DELETE);
            ByteHelper.writeClass(os, t);
            ByteHelper.writeKey(os, k);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public void close() throws IsamException {
        log.fine("Close");
        try {
            os.write(IsamProtocol.CLOSE);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
            log.finest("Closing streams");
            os.close();
            is.close();
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public <T, TK> IsamResult<T> readXX(int cmd, Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadXX class " + t.getName() + " with key " + k);
        try {
            os.write(cmd);
            ByteHelper.writeClass(os, t);
            ByteHelper.writeKey(os, k);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
            return new RemoteIsamResult<T>(this);
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public boolean resultRead() throws IsamException, RecordException {
        log.fine("Result read");
        try {
            os.write(IsamProtocol.RESULT_READ);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
            return is.read() != 0;
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    @SuppressWarnings("unchecked")
    public <T> T resultCurrent() throws IsamException, RecordException {
        log.fine("Result current");
        try {
            os.write(IsamProtocol.RESULT_CURRENT);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
            return (T)ByteHelper.readObject(is, log);
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Remote exception", e);
        }
    }
    public void resultClose() throws IsamException {
        log.fine("Result close");
        try {
            os.write(IsamProtocol.RESULT_CLOSE);
            os.flush();
            log.finer("Command send to server");
            int ok = is.read();
            if(ok != IsamProtocol.OK) {
                log.fine("Not OK");
                throw new IsamException("Remote server exception", ByteHelper.readException(is));
            } else {
                log.fine("OK");
            }
        } catch (IOException e) {
            throw new IsamException("Remote exception", e);
        }
    }
}
